#include <stdint.h>
#include <iostream>
#include <sys/time.h>

#include "chr_chronos.h"

#define TEXT_LEN 80

typedef struct _key_st {
  uint32_t id;
} key_st;

typedef struct _value_st {
  uint64_t timestamp;
  char text[TEXT_LEN];
} value_st;

using namespace std;

/* convert a key_st sruct from big endian to host format */
void betoh(key_st * key) {
  key->id = be32toh(key->id);
}

/* convert a key_st sruct from host format to big endian */
void htobe(key_st * key) {
  key->id = htobe32(key->id);
}

/* get the number of milliseconds since 1970-01-01 00:00:00.000 */
uint64_t get_current_timestamp() {
  struct timeval time;
  gettimeofday(&time, NULL);
  return ((uint64_t)time.tv_sec) * 1000 + time.tv_usec / 1000;
}

/* insert a key-value pair (key:id, value:timestamp,text) or update the value if the key exists */
void insert_or_update(chr_cursor * cur, uint32_t id, char * text) {
  /* set the key */
  key_st key;
  key.id = id;
  htobe(&key);
  /* set the value */
  value_st value;
  value.timestamp = get_current_timestamp();
  strcpy(value.text, text);
  /* insert the pair */
  int ret = cur->insert((unsigned char *)&key, (unsigned char *)&value);

  /* if exists, update it */
  if (ret == CHR_DUPKEY) {
    printf("Updating text n°%d\n", id);
    cur->update((unsigned char *)&key, (unsigned char *)&value);
  }
}

/* display all key-value pairs in the table */
void display_contents(chr_table_ctx * ctx) {
  cout << "Table's content:" << endl;
  key_st key;
  value_st value;
  bool empty = true;
  /* open a cursor on the first pair */
  chr_cursor * read_cur = ctx->open_cursor(NULL, CHR_FIRST);
  /* while there is a pair to be read */
  while (read_cur->read_next((unsigned char *)&key, (unsigned char *)&value) != CHR_EOF) {
    empty = false;
    /* convert the key */
    betoh(&key);
    /* convert the timestamp to a readable format */
    char buffer[24];
    time_t epoch = value.timestamp / 1000;
    struct tm * timeinfo = localtime(&epoch);
    strftime(buffer, 20, "%Y-%m-%d %H:%M:%S", timeinfo);
    unsigned int ms = value.timestamp % 1000;
    sprintf(buffer, "%s.%03u", buffer, ms);
    /* display the pair */
    cout << "Text n°" << key.id << " (" << buffer << "): " << value.text << endl;
  }
  if (empty) {
    cout << "<empty>" << endl;
  }
  /* close the cursor */
  ctx->close_cursor(read_cur);
}

int main(int argc, const char * argv[]) {
  /* init the database */
  chr_chronos * chronos = new chr_chronos("/tmp/chronos_dummy_flash_device",
      "/tmp/chronos_dummy_persistence_folder/");

  /* open the table */
  chr_table_ctx * texts_ctx = chronos->open_table("texts");
  /* create the table if it does not exist (open failed) */
  if (texts_ctx == NULL) {
    cout << "Table \"texts\" does not exist, creating it.\n"
         << "key: id INT32\n"
         << "value: timestamp INT64, text CHAR(" << TEXT_LEN << ")" << endl;
    chronos->create_table("texts", sizeof(key_st), sizeof(value_st));
    texts_ctx = chronos->open_table("texts");
  }
  
  chr_cursor * insert_cur = texts_ctx->open_cursor(NULL, CHR_FIRST);

  int id = 0;
  
  while (id != -1) {
    /* display the table's contents */
    display_contents(texts_ctx);
  
    /* get input from user */
    printf("New text id: (-1 to quit, -2 to generate sequences) ");
    char buffer[20];
    fgets(buffer, sizeof(buffer), stdin);
    sscanf(buffer, "%d", &id);
    if (id == -2) {
      int start_id = -1;
      int end_id = -1;
      while (start_id < 0) {
        printf("Start id (integer >= 0): ");
        fgets(buffer, sizeof(buffer), stdin);
        sscanf(buffer, "%d", &start_id);
      }
      while (end_id <= start_id) {
        printf("End id (integer > %u): ", start_id);
        fgets(buffer, sizeof(buffer), stdin);
        sscanf(buffer, "%d", &end_id);
      } 
      for (int i = start_id; i <= end_id; i++) {
        /* insert the new value */
        char text[TEXT_LEN];
        sprintf(text, "Placeholder text n°%d", i);
        insert_or_update(insert_cur, i, text);
      }
    } else if (id >= 0) {
      char text[TEXT_LEN];
      printf("Text contents: ");
      fgets(text, TEXT_LEN, stdin);
      /* trim trailing newline */
      if (text[strlen(text) - 1] == '\n')
      {
        text[strlen(text) - 1] = '\0';
      }
      printf("New text n°%d: %s\n", id, text);

      /* insert the new value */
      insert_or_update(insert_cur, id, text);
    } else if (id != -1) {
      cout << "wrong id: " << id << endl;
    }
  }
    
  /* close cursors, tables and databases */
  texts_ctx->close_cursor(insert_cur);
  chronos->close_table(texts_ctx);
  delete chronos;
}

