#include "chr_chronos.h"

#include <stdlib.h>
#include <string.h>

#include "chr_common.h"
#include "chr_flashgui_text.h"
//#include "chr_flashgui_gtk.h"

/**
 * Opens or initializes a database. Persistance files in {@link folder}
 * are not mandatory to rebuild the database.
 *
 * @param device url of the file or raw device to keep the database on
 * @param folder url of the folder to keep persistance files on
 */
chr_chronos::chr_chronos(const char * device, const char * folder)
{
  this->folder = (char *)malloc(strlen(folder) + 1);
  strcpy(this->folder, folder);

  std::string chr_file = folder;
  chr_file.append(CHR_FILE);
  FILE * chr_fd = fopen(chr_file.c_str(), "r");
  std::string chr_flash_file = folder;
  chr_flash_file.append(CHR_FLASH_FILE);
  this->flashgui = new CHR_FLASH_GUI();
  this->flash = new chr_flash(device, chr_flash_file.c_str(), flashgui);
  if(chr_fd != NULL)
  {
    int c = fgetc(chr_fd);
    while(c != EOF)
    {
      std::string table_name;
      while(c != '\0')
      {
        table_name.append(1, (char)c);
        c = fgetc(chr_fd);
      }
      chr_table_ctx * ctx = (chr_table_ctx *)
        malloc(sizeof(chr_table_ctx));
      if(fread(&(ctx->key_size), sizeof(size_t), 1, chr_fd) != 1)
      {
        printf("[file %s, line %d]: fread failed\n",
          __FILE__, __LINE__);
      }
      if(fread(&(ctx->value_size), sizeof(size_t), 1, chr_fd) != 1)
      {
        printf("[file %s, line %d]: fread failed\n",
          __FILE__, __LINE__);
      }
      chr_lsn root_lsn;
      if(fread(&root_lsn, sizeof(chr_lsn), 1, chr_fd) != 1)
      {
        printf("[file %s, line %d]: fread failed\n",
          __FILE__, __LINE__);
      }
      ctx->table = new chr_table(ctx->key_size, ctx->value_size, flash,
          root_lsn);
      ctx->open_count = 0;
      tables[table_name] = ctx;
      c = fgetc(chr_fd);
    }
    fclose(chr_fd);
  }
}

/**
 * Closes a database.
 */
chr_chronos::~chr_chronos()
{
  std::string chr_file = folder;
  chr_file.append(CHR_FILE);
  FILE * chr_fd = fopen(chr_file.c_str(), "w+");
  if(chr_fd == NULL)
  {
    fprintf(stderr, "error: can not open %s\n", chr_file.c_str());
    fprintf(stderr, "tables will not be saved\n");
  }
  else
  {
    std::map<std::string, chr_table_ctx *>::iterator it = tables.begin();
    while(it != tables.end())
    {
      std::pair<std::string, chr_table_ctx *> p = *it; 
      chr_table_ctx * ctx = p.second;
      if(ctx == NULL)
      {
        fprintf(stderr, "unknow error while closing: data might be lost\n");
        fprintf(stderr, "this might be caused by wrong table names\n");
      } else {
        if(ctx->open_count > 0)
        {
          fprintf(stderr,
              "error: can not flush table %s, table is still opened\n",
              p.first.c_str());
          fprintf(stderr, "error: data might be lost\n");
        }
        chr_lsn root_lsn = ctx->table->close();
        delete ctx->table;
        fwrite(p.first.c_str(), sizeof(char), p.first.length() + 1,
            chr_fd);
        fwrite(&(ctx->key_size), sizeof(size_t), 1, chr_fd);
        fwrite(&(ctx->value_size), sizeof(size_t), 1, chr_fd);
        fwrite(&root_lsn, sizeof(chr_lsn), 1, chr_fd);
      }
      tables.erase(it);
      free(ctx);
      it = tables.begin();
    }
    fclose(chr_fd);
  }
  /* close all tables */
  free(folder);
  delete this->flash;
  delete this->flashgui;
}

/**
 * Creates a new table.
 *
 * @param  table_name name of the new table
 * @param  key_size   size of the key (in bytes)
 * @param  value_size size of the value (in bytes)
 * @return CHR_ERR if the table already exists,
 *         CHR_OK otherwise
 */
int chr_chronos::create_table(const char * table_name, size_t key_size,
    size_t value_size)
{
  if(tables[table_name] != NULL)
  {
    return(CHR_ERR);
  }

  chr_table_ctx * ctx = (chr_table_ctx *)
    malloc(sizeof(chr_table_ctx));
  ctx->key_size = key_size;
  ctx->value_size = value_size;
  ctx->open_count = 0;
  ctx->table = new chr_table(ctx->key_size, ctx->value_size, flash);
  tables[table_name] = ctx;
  return(CHR_OK);
}

/**
 * Removes a table from the database. Deletes its content.
 *
 * @param  table_name name of the table
 * @return CHR_NOTFOUND if the table doesn't exist,
 *         CHR_ERR if the table is not closed,
 *         CHR_OK otherwise
 */
int chr_chronos::drop_table(const char * table_name)
{
  chr_table_ctx * ctx = tables[table_name];
  if(ctx == NULL)
  {
    return(CHR_NOTFOUND);
  }
  if(ctx->open_count > 0)
  {
    fprintf(stderr,
        "error: can not drop table %s, table is still opened\n",
        table_name);
    return(CHR_ERR);
  }
  ctx->table->drop();
  delete ctx->table;
  tables.erase(table_name);
  free(ctx);
  return(CHR_OK);
}

/**
 * Opens an existing table.
 *
 * @param  table_name name of the table
 * @return NULL if the table doesn't exist,
 *         a valid table handler otherwise
 */
chr_table_ctx * chr_chronos::open_table(const char * table_name)
{
  chr_table_ctx * ctx = tables[table_name];
  if(ctx == NULL)
  {
    return(NULL);
  }
  ctx->open_count++;
  if(ctx->table == NULL)
  {
    ctx->table = new chr_table(ctx->key_size, ctx->value_size, flash);
  }
  return(ctx);
}

/**
 * Closes an opened table.
 *
 * @param  ctx a table handler previously opened
 * @return CHR_ERR if the handler is not valid,
 *         CHR_OK otherwise
 */
int chr_chronos::close_table(chr_table_ctx * ctx)
{
  if(ctx == NULL)
  {
    return(CHR_ERR);
  }
  ctx->open_count--;
  return(CHR_OK);
}

/**
 * Opens a new cursor.
 *
 * @param  key  byte buffer containing the key,
 *              NULL if {@link type} is CHR_FIRST or CHR_LAST
 * @param  type position of the cursor,
 *              CHR_KEY to position on a given {@link key},
 *              CHR_FIRST to position on the first tuple,
 *              CHR_LAST to position on the last tuple
 * @return a cursor handler
 */
chr_cursor * chr_table_ctx::open_cursor(const unsigned char * key, chr_pos_type type)
{
  return(table->open_cursor(key, type));
}

/**
 * Closes an opened cursor.
 *
 * @param  cursor a cursor handler
 * @return CHR_OK
 */
int chr_table_ctx::close_cursor(chr_cursor * cursor)
{
  delete cursor;
  return(CHR_OK);
}

