#include "chr_inode.h"

#include <stdlib.h>
#include <assert.h>

#include "chr_parse.h"
#include "chr_leaf.h"

chr_inode::chr_inode(const size_t key_size, chr_inode * parent,
    chr_table * table, chr_flash * flash, bool kmininf,
    const unsigned char * kmin, bool kmaxinf,
    const unsigned char * kmax, chr_lsn lsn,
    const unsigned char * buf)
  : chr_node(key_size, chr_parse::keys_per_inode(key_size),
      parent, table, flash, kmininf, kmin, kmaxinf, kmax),
    dirty(false),
    lsn(lsn),
    ptrs(NULL)
{
  ptrs = (chr_ptr *)malloc((keys_per_node + 1) * sizeof(chr_ptr));
  if(ptrs == NULL)
  {
    throw std::bad_alloc();
  }

  if(lsn != NULL_LSN)
  {
    /* load from flash */
    assert(buf != NULL);
    assert(!(chr_parse::is_leaf(buf)));
    nb_keys = chr_parse::inode_nb_keys(buf);
    for(unsigned int i = 0; i<nb_keys; i++)
    {
      memcpy(key(i), chr_parse::inode_key(buf, i, key_size), key_size);
      ptrs[i] = chr_parse::inode_ptr(buf, i, key_size);
    }
    ptrs[nb_keys] = chr_parse::inode_ptr(buf, nb_keys, key_size);
  }
  else
  {
    nb_keys = 0;
    /* in this case, init_son should be called after this constructor */
  }
  if(parent != NULL)
  {
    parent->bind();
  }
}

chr_inode::~chr_inode()
{
  if(dirty)
  {
    /* flush */
    unsigned char * buf = (unsigned char *)malloc(CHR_BLOCK_SIZE);
    to_sector(buf);
    if(lsn != NULL_LSN)
    {
      flash->trim(lsn);
    }
    flash->nameless_write(&lsn, buf);
    free(buf);
  }
  if(parent != NULL)
  {
    parent->update(key(0), lsn);
  }
  else
  {
    table->root = NULL;
    table->root_lsn = lsn;
  }
  free(ptrs);
  if(parent != NULL)
  {
    /* update parent ptr */
    parent->unbind(); /* unbind should be the last call */
  }
}

void chr_inode::to_sector(unsigned char * buf)
{
  chr_parse::inode_to_sector(buf, nb_keys, key_size, keys, ptrs);
}

void chr_inode::init_son(const chr_ptr ptr)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  assert(nb_keys == 0);
  assert(ptr.is_memory() || ptr.get_lsn() != NULL_LSN);
  ptrs[0] = ptr;
  dirty = true;
}

void chr_inode::drop()
{
  assert(parent == NULL || parent->son(key(0)) == this);
  bool leaf_sons = false;
  if(!ptrs[0].is_memory())
  {
    unsigned char * buf = (unsigned char *)malloc(CHR_BLOCK_SIZE);
    flash->read(ptrs[0].get_lsn(), buf);
    if(chr_parse::is_leaf(buf))
    {
      leaf_sons = true;
    }
    free(buf);
  }
  else
  {
    if(ptrs[0].get_node()->is_leaf())
    {
      leaf_sons = true;
    }
  }
  for(unsigned int i=0; i <= nb_keys; i++)
  {
    if(leaf_sons)
    {
      if(ptrs[i].is_memory())
      {
        ptrs[i].get_node()->leaf()->drop();
        /* flash->trim(ptrs[i].get_node()->leaf()->lsn); */
      }
      else
      {
        if(ptrs[i].get_lsn() == NULL_LSN)
        {
          fprintf(stderr, "error: son address invalid\n");
        }
        else
        {
          flash->trim(ptrs[i].get_lsn());
        }
      }
    }
    else
    {
      if(!(ptrs[i].is_memory()))
      {
        load_ptr(i);
      }
      ptrs[i].get_node()->inode()->drop();
    }
  }
  if(lsn != NULL_LSN)
  {
    flash->trim(lsn);
  }
  parent = NULL;
  dirty = false;
  delete this;
}

void chr_inode::load_ptr(unsigned int i)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  assert(i <= nb_keys);
  assert(!(ptrs[i].is_memory()));
  unsigned char * buf = (unsigned char *)malloc(CHR_BLOCK_SIZE);
  assert(buf != NULL);
  flash->read(ptrs[i].get_lsn(), buf);
  bool sonkmininf;
  unsigned char * sonkmin = NULL;
  bool sonkmaxinf;
  unsigned char * sonkmax = NULL;
  if(i == 0)
  {
    sonkmininf = kmininf;
    sonkmin = kmin;
  }
  else
  {
    sonkmininf = false;
    sonkmin = key(i - 1);
  }
  if(i == nb_keys)
  {
    sonkmaxinf = kmaxinf;
    sonkmax = kmax;
  }
  else
  {
    sonkmaxinf = false;
    sonkmax = key(i);
  }
  if(chr_parse::is_leaf(buf))
  {
    chr_leaf * son = new chr_leaf(key_size, table->get_value_size(), this,
        table, flash, sonkmininf, sonkmin, sonkmaxinf, sonkmax,
        ptrs[i].get_lsn(), buf);
    ptrs[i] = son;
  }
  else
  {
    chr_inode * son = new chr_inode(key_size, this, table, flash,
        sonkmininf, sonkmin, sonkmaxinf, sonkmax, ptrs[i].get_lsn(), buf);
    ptrs[i] = son;
  }
  free(buf);
}

int chr_inode::insert_left(const unsigned char * k, const chr_ptr ptr)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  assert(nb_keys < keys_per_node);
  assert(in_range(k));
  assert(ptr.is_memory() || ptr.get_lsn() != NULL_LSN);
  int pos = nb_keys - 1;
  while(pos >= 0 && comp(key(pos), k) > 0)
  {
    memcpy(key(pos + 1), key(pos), key_size);
    ptrs[pos + 2] = ptrs[pos + 1];
    pos--;
  }
  assert(pos < 0 || comp(key(pos), k) != 0);
  pos++;
  memcpy(key(pos), k, key_size);
  ptrs[pos + 1] = ptrs[pos];
  ptrs[pos] = ptr;
  nb_keys++;
  dirty = true;
  if(nb_keys == keys_per_node)
  {
    split();
  }
  return(CHR_OK);
}

int chr_inode::insert_right(const unsigned char * k, const chr_ptr ptr)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  assert(nb_keys < keys_per_node);
  assert(in_range(k));
  assert(ptr.is_memory() || ptr.get_lsn() != NULL_LSN);
  int pos = nb_keys - 1;
  while(pos >= 0 && comp(key(pos), k) > 0)
  {
    memcpy(key(pos + 1), key(pos), key_size);
    ptrs[pos + 2] = ptrs[pos + 1];
    pos--;
  }
  assert(pos < 0 || comp(key(pos), k) != 0);
  pos++;
  memcpy(key(pos), k, key_size);
  ptrs[pos + 1] = ptr;
  nb_keys++;
  dirty = true;
  if(nb_keys == keys_per_node)
  {
    split();
  }
  return(CHR_OK);
}

int chr_inode::split()
{
  assert(parent == NULL || parent->son(key(0)) == this);
  /* print(print_k, print_v); */
  assert(nb_keys == keys_per_node);
  if(parent == NULL)
  {
    /* allocate new root */
    parent = new chr_inode(key_size, NULL, table, flash,
        true, NULL, true, NULL, NULL_LSN, NULL);
    assert(parent != NULL);
    parent->init_son(this);
    /* bind twice, for this node and for being root */
    parent->bind();
    parent->bind();
    /* unbind previous root (this) */
    this->unbind();
    table->root = parent;
  }

  int split_pos = keys_per_node/2;
  unsigned char * split_key = (unsigned char *)malloc(key_size);
  assert(split_key != NULL);
  memcpy(split_key, key(split_pos), key_size);
  chr_inode * brother = new chr_inode(key_size, parent, table, flash,
      kmininf, kmin, false, split_key, NULL_LSN, NULL);
  set_kmin(false, split_key);
  brother->bind();
  /* copy first keys - ptrs to brother */
  for(int i=0; i<split_pos; i++)
  {
    memcpy(brother->key(i), this->key(i), key_size);
    brother->ptrs[i] = this->ptrs[i];
  }
  brother->ptrs[split_pos] = this->ptrs[split_pos];
  brother->nb_keys = split_pos;
  brother->dirty = true;

  /* update current inode (this) */
  for(unsigned int i=0; i<keys_per_node - split_pos - 1; i++)
  {
    /* TODO: one big memcpy */
    memcpy(this->key(i), this->key(split_pos + 1 + i), key_size);
    ptrs[i] = ptrs[split_pos + 1 + i];
  }
  ptrs[keys_per_node - split_pos - 1] = ptrs[keys_per_node];
  nb_keys = keys_per_node - split_pos - 1;

  /* update bindings */
  this->bind();
  for(unsigned int i=0; i<=brother->nb_keys; i++)
  {
    if(brother->ptrs[i].is_memory())
    {
      brother->ptrs[i].get_node()->parent = brother;
      brother->bind();
      this->unbind();
    }
  }

  parent->insert_left(split_key, brother);
  free(split_key);
  brother->unbind();
  this->unbind();
  return(CHR_OK);
}

int chr_inode::del(const unsigned char * k)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  int pos;
  search(&pos, k);
  if(pos == -1)
  {
    ptrs[0] = ptrs[1];
    pos = 0;
  }
  while(pos < (int)nb_keys)
  {
    memcpy(key(pos), key(pos + 1), key_size);
    ptrs[pos + 1] = ptrs[pos + 2];
    pos++;
  }
  nb_keys--;
  dirty = true;
  return(CHR_OK);
}

chr_node * chr_inode::son(const unsigned char * k)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  int pos;
  search(&pos, k);
  assert(pos <= (int)nb_keys);
  if (!ptrs[pos + 1].is_memory())
  {
    load_ptr(pos + 1);
  }
  return(ptrs[pos + 1].get_node());
}

chr_node * chr_inode::nextson(const unsigned char * k)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  int pos;
  search(&pos, k);
  assert(pos <= (int)nb_keys);
  if(pos >= (int)nb_keys - 1)
  {
    return(NULL);
  }
  else
  {
    if(!(ptrs[pos + 2].is_memory()))
    {
      load_ptr(pos + 2);
    }
    return(ptrs[pos + 2].get_node());
  }
}

chr_node * chr_inode::prevson(const unsigned char * k)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  int pos;
  search(&pos, k);
  assert(pos <= (int)nb_keys);
  if(pos == -1)
  {
    return(NULL);
  }
  else
  {
    if(!(ptrs[pos].is_memory()))
    {
      load_ptr(pos);
    }
    return(ptrs[pos].get_node());
  }
}

int chr_inode::update(const unsigned char * k, const chr_ptr ptr)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  assert(ptr.is_memory() || ptr.get_lsn() != NULL_LSN);
  int pos;
  search(&pos, k);
  assert(pos <= (int)nb_keys);
  ptrs[pos + 1] = ptr;
  dirty = true;
  return(CHR_OK);
}

chr_leaf * chr_inode::search_leaf(const unsigned char * k,
    chr_pos_type type, bool minusinf, const unsigned char * kmin,
    bool plusinf, const unsigned char * kmax)
{
  assert(parent == NULL || parent->son(key(0)) == this);
  int pos;
  if(type == CHR_KEY)
  {
    search(&pos, k);
    pos++;
  }
  else if(type == CHR_FIRST)
  {
    pos = 0;
  }
  else
  {
    pos = nb_keys;
  }
  if(pos > 0)
  {
    minusinf = false;
    kmin = key(pos - 1);
  }
  if((unsigned int)pos < nb_keys)
  {
    plusinf = false;
    kmax = key(pos);
  }
  if(!(ptrs[pos].is_memory()))
  {
    load_ptr(pos);
  }
  chr_node * node = ptrs[pos].get_node();
  if(node->is_leaf())
  {
    node->leaf()->update_range(minusinf, kmin, plusinf, kmax);
    return(node->leaf());
  }
  else
  {
    return(node->inode()->search_leaf(k, type,
          minusinf, kmin, plusinf, kmax));
  }
}

inline bool chr_inode::is_leaf()
{
  return(false);
}

inline chr_leaf * chr_inode::leaf()
{
  /* this is not a leaf */
  assert(false);
  return(NULL);
}

inline chr_inode * chr_inode::inode()
{
  return(this);
}

void chr_inode::get_state(chr_lsn * lsn, bool * kmininf,
    unsigned char ** kmin, bool * kmaxinf, unsigned char ** kmax,
    size_t * key_size, unsigned int * nb_keys, unsigned char ** keys,
    chr_ptr ** ptrs)
{
  *lsn = this->lsn;
  *kmininf = this->kmininf;
  *kmin = this->kmin;
  *kmaxinf = this->kmaxinf;
  *kmax = this->kmax;
  *key_size = this->key_size;
  *nb_keys = this->nb_keys;
  *keys = this->keys;
  *ptrs = this->ptrs;
}

void chr_inode::dbug_check(bool kmininf, const unsigned char * kmin,
    bool kmaxinf, const unsigned char * kmax)
{
  /* kmin <= ki < ki+1 < kmax */
  if(nb_keys == 0)
  {
    assert(kmininf || kmaxinf || comp(kmin, kmax) < 0);
  }
  else
  {
    assert(kmininf || comp(kmin, key(0)) <= 0);
    assert(kmaxinf || comp(key(nb_keys - 1), kmax) < 0);
    for(unsigned int i = 0; i < nb_keys - 1; i++)
    {
      assert(comp(key(i), key(i + 1)) < 0);
      if(ptrs[i].is_memory())
      {
        ptrs[i].get_node()->dbug_check(kmininf, kmin, kmaxinf, kmax);
      }
      else
      {
        /* bug?
        load_ptr(i);
        ptrs[i].get_node()->dbug_check(kmininf, kmin, kmaxinf, kmax);
        delete ptrs[i].get_node();
        */
      }
    }
  }
}

