#include "chr_table.h"

#include <stdlib.h>
#include <sstream>

#include "chr_parse.h"
#include "chr_inode.h"
#include "chr_leaf.h"
#include "chr_cursor.h"
#include "chr_btreegui_latex.h"

chr_table::chr_table(const size_t key_size, const size_t value_size,
    chr_flash * flash, const chr_lsn root_lsn)
  : key_size(key_size), value_size(value_size), flash(flash),
    root_lsn(root_lsn)
{
  btreegui = new CHR_BTREE_GUI();
  btreegui_dbug = new chr_btreegui_latex();
  if(root_lsn == NULL_LSN)
  {
    /* new tree */
    root = new chr_leaf
      (
        key_size,
        value_size,
        NULL,
        this,
        flash,
        true,
        NULL,
        true,
        NULL,
        NULL_LSN,
        NULL
      );
    }
  else
  {
    /* load tree */
    unsigned char * buf = (unsigned char *)malloc(CHR_BLOCK_SIZE);
    flash->read(root_lsn, buf);
    if(chr_parse::is_leaf(buf))
    {
      root = new chr_leaf
        (
          key_size,
          value_size,
          NULL,
          this,
          flash,
          true,
          NULL,
          true,
          NULL,
          root_lsn,
          buf
        );
    }
    else
    {
      root = new chr_inode
        (
          key_size,
          NULL,
          this,
          flash,
          true,
          NULL,
          true,
          NULL,
          root_lsn,
          buf
        );
    }
    free(buf);
  }
  root->bind();
}

chr_table::~chr_table()
{
  /* close() should be called before */
  assert(root == NULL);
  delete btreegui;
  delete btreegui_dbug;
}

int chr_table::drop()
{
  assert(root != NULL);
  if(root->bind_count > 1)
  {
    fprintf(stderr,
        "can not drop table, cursors are probably still opened\n");
    return(CHR_ERR);
  }
  assert(root->bind_count == 1);
  root->drop();
  return(CHR_OK);
}

chr_lsn chr_table::close()
{
  root->unbind();
  if(root != NULL)
  {
    fprintf(stderr,
        "can not close table, cursors are probably still opened\n");
    root->bind();
    return(NULL_LSN);
  }
  return(root_lsn);
}

chr_leaf * chr_table::search(const unsigned char * key, chr_pos_type type)
{
  /* at the same time, we update the range of each accessed node */
  if(root->is_leaf())
  {
    root->leaf()->update_range(true, NULL, true, NULL);
    return(root->leaf());
  }
  else
  {
    return(root->inode()->search_leaf(key, type, true, NULL, true, NULL));
  }
}

chr_cursor * chr_table::open_cursor(const unsigned char * key,
    chr_pos_type type)
{
  chr_cursor * cur = new chr_cursor(this, key_size, value_size);
  cur->move(key, type);
  return(cur);
}

size_t chr_table::get_value_size()
{
  return(this->value_size);
}

size_t chr_table::get_key_size()
{
  return(this->key_size);
}

const chr_flash * chr_table::get_flash()
{
  return(this->flash);
}

void chr_table::dbug_print()
{
  btreegui_dbug->print_tree(root);
}

void chr_table::dbug_check()
{
  root->dbug_check(true, NULL, true, NULL);
}

#ifndef CHR_NO_GUI
void chr_table::print()
{
  btreegui->print_tree(root);
}
#endif /* CHR_NO_GUI */

